package main

import (
	"encoding/base64"
	"flag"
	"fmt"
	"image"
	_ "image/gif"
	_ "image/jpeg"
	_ "image/png"
	"log"
	"math"
	"os"
	"path/filepath"
	"strings"
	"time"
)

const version = 0.1

type RGB struct {
	R byte
	G byte
	B byte
}

type C64RGB struct {
	name       string
	colorindex byte
	rgb        RGB
}

type Screen struct {
	charcolors [1000]map[RGB]byte
}

type Koala struct {
	bitmap      [8000]byte
	screencolor [1000]byte
	d800color   [1000]byte
	bgcolor     [1]byte
}

// binary blob containing c64 startaddress + basicstart + koala display code
const koaladisplayb64 = `AQgLCAoAnjIwNjIAAAAAeKk3hQGpO40R0KkYjRjQqdiNFtCtEEeNINCNIdCiAL1AP50ABL1AQJ0A
Bb1AQZ0ABr1AQp0AB70oQ50A2L0oRJ0A2b0oRZ0A2r0oRp0A2+jQzUxgCA==`

var koaladisplay []byte
var helpbool bool
var quiet bool
var verbose bool
var display bool
var c64palettes = map[string][]C64RGB{
	"vice": []C64RGB{
		{"black", 0, RGB{0x00, 0x00, 0x00}},
		{"white", 1, RGB{0xff, 0xff, 0xff}},
		{"red", 2, RGB{0xbc, 0x52, 0x41}},
		{"cyan", 3, RGB{0x8f, 0xef, 0xfb}},
		{"purple", 4, RGB{0xb9, 0x56, 0xeb}},
		{"green", 5, RGB{0x7e, 0xdb, 0x40}},
		{"blue", 6, RGB{0x55, 0x3f, 0xe4}},
		{"yellow", 7, RGB{0xff, 0xff, 0x77}},
		{"orange", 8, RGB{0xc1, 0x7b, 0x1d}},
		{"brown", 9, RGB{0x82, 0x63, 0x00}},
		{"lightred", 10, RGB{0xf4, 0x94, 0x86}},
		{"darkgrey", 11, RGB{0x72, 0x72, 0x72}},
		{"grey", 12, RGB{0xa4, 0xa4, 0xa4}},
		{"lightgreen", 13, RGB{0xcd, 0xff, 0x98}},
		{"lightblue", 14, RGB{0x9e, 0x8d, 0xff}},
		{"lightgrey", 15, RGB{0xd5, 0xd5, 0xd5}},
	},
	"gimp": []C64RGB{
		{"black", 0, RGB{0x00, 0x00, 0x00}},
		{"white", 1, RGB{0xff, 0xff, 0xff}},
		{"red", 2, RGB{0x68, 0x37, 0x2b}},
		{"cyan", 3, RGB{0x70, 0xa4, 0xb2}},
		{"purple", 4, RGB{0x6f, 0x3d, 0x86}},
		{"green", 5, RGB{0x58, 0x8d, 0x43}},
		{"blue", 6, RGB{0x35, 0x28, 0x79}},
		{"yellow", 7, RGB{0xb8, 0xc7, 0x6f}},
		{"orange", 8, RGB{0x6f, 0x4f, 0x25}},
		{"brown", 9, RGB{0x43, 0x39, 0x00}},
		{"lightred", 10, RGB{0x9a, 0x67, 0x59}},
		{"darkgrey", 11, RGB{0x44, 0x44, 0x44}},
		{"grey", 12, RGB{0x6c, 0x6c, 0x6c}},
		{"lightgreen", 13, RGB{0x9a, 0xd2, 0x84}},
		{"lightblue", 14, RGB{0x6c, 0x5e, 0xb5}},
		{"lightgrey", 15, RGB{0x95, 0x95, 0x95}},
	},
	"pantaloon": []C64RGB{
		{"black", 0, RGB{0, 0, 0}},
		{"white", 1, RGB{255, 255, 255}},
		{"red", 2, RGB{104, 55, 43}},
		{"cyan", 3, RGB{131, 240, 220}},
		{"purple", 4, RGB{111, 61, 134}},
		{"green", 5, RGB{89, 205, 54}},
		{"blue", 6, RGB{65, 55, 205}},
		{"yellow", 7, RGB{184, 199, 111}},
		{"orange", 8, RGB{209, 127, 48}},
		{"brown", 9, RGB{67, 57, 0}},
		{"lightred", 10, RGB{154, 103, 89}},
		{"darkgrey", 11, RGB{91, 91, 91}},
		{"grey", 12, RGB{142, 142, 142}},
		{"lightgreen", 13, RGB{157, 255, 157}},
		{"lightblue", 14, RGB{117, 161, 236}},
		{"lightgrey", 15, RGB{193, 193, 193}},
	},
	"archmage": []C64RGB{
		{"black", 0, RGB{0x00, 0x00, 0x00}},
		{"white", 1, RGB{0xff, 0xff, 0xff}},
		{"red", 2, RGB{0x89, 0x40, 0x36}},
		{"cyan", 3, RGB{0x7a, 0xbf, 0xc7}},
		{"purple", 4, RGB{0x8a, 0x46, 0xae}},
		{"green", 5, RGB{0x68, 0xa9, 0x41}},
		{"blue", 6, RGB{0x3e, 0x31, 0xa2}},
		{"yellow", 7, RGB{0xd0, 0xdc, 0x71}},
		{"orange", 8, RGB{0x90, 0x5f, 0x25}},
		{"brown", 9, RGB{0x5c, 0x47, 0x00}},
		{"lightred", 10, RGB{0xbb, 0x77, 0x6d}},
		{"darkgrey", 11, RGB{0x55, 0x55, 0x55}},
		{"grey", 12, RGB{0x80, 0x80, 0x80}},
		{"lightgreen", 13, RGB{0xac, 0xea, 0x88}},
		{"lightblue", 14, RGB{0x7c, 0x70, 0xda}},
		{"lightgrey", 15, RGB{0xab, 0xab, 0xab}},
	},
	"electric cocillana": []C64RGB{
		{"black", 0, RGB{0x00, 0x00, 0x00}},
		{"white", 1, RGB{0xff, 0xff, 0xff}},
		{"red", 2, RGB{0x8b, 0x1f, 0x00}},
		{"cyan", 3, RGB{0x6f, 0xdf, 0xb7}},
		{"purple", 4, RGB{0xa7, 0x3b, 0x9f}},
		{"green", 5, RGB{0x4a, 0xb5, 0x10}},
		{"blue", 6, RGB{0x08, 0x00, 0x94}},
		{"yellow", 7, RGB{0xf3, 0xeb, 0x5b}},
		{"orange", 8, RGB{0xa5, 0x42, 0x00}},
		{"brown", 9, RGB{0x63, 0x29, 0x18}},
		{"lightred", 10, RGB{0xcb, 0x7b, 0x6f}},
		{"darkgrey", 11, RGB{0x45, 0x44, 0x44}},
		{"grey", 12, RGB{0x9f, 0x9f, 0x9f}},
		{"lightgreen", 13, RGB{0x94, 0xff, 0x94}},
		{"lightblue", 14, RGB{0x4a, 0x94, 0xd6}},
		{"lightgrey", 15, RGB{0xbd, 0xbd, 0xbd}},
	},
	"ste": []C64RGB{
		{"black", 0, RGB{0x00, 0x00, 0x00}},
		{"white", 1, RGB{0xff, 0xff, 0xff}},
		{"red", 2, RGB{0xc8, 0x35, 0x35}},
		{"cyan", 3, RGB{0x83, 0xf0, 0xdc}},
		{"purple", 4, RGB{0xcc, 0x59, 0xc6}},
		{"green", 5, RGB{0x59, 0xcd, 0x36}},
		{"blue", 6, RGB{0x41, 0x37, 0xcd}},
		{"yellow", 7, RGB{0xf7, 0xee, 0x59}},
		{"orange", 8, RGB{0xd1, 0x7f, 0x30}},
		{"brown", 9, RGB{0x91, 0x5f, 0x33}},
		{"lightred", 10, RGB{0xf9, 0x9b, 0x97}},
		{"darkgrey", 11, RGB{0x5b, 0x5b, 0x5b}},
		{"grey", 12, RGB{0x8e, 0x8e, 0x8e}},
		{"lightgreen", 13, RGB{0x9d, 0xff, 0x9d}},
		{"lightblue", 14, RGB{0x75, 0xa1, 0xec}},
		{"lightgrey", 15, RGB{0xc1, 0xc1, 0xc1}},
	},
}

func init() {
	flag.BoolVar(&quiet, "q", false, "")
	flag.BoolVar(&quiet, "quiet", false, "quiet")
	flag.BoolVar(&verbose, "v", false, "")
	flag.BoolVar(&verbose, "verbose", false, "verbose")
	flag.BoolVar(&display, "d", false, "")
	flag.BoolVar(&display, "display", false, "include koala displayer")
	flag.BoolVar(&helpbool, "h", false, "")
	flag.BoolVar(&helpbool, "help", false, "help")
}

func printusage() {
	fmt.Println("usage: ./png2prg [-d -display -q -quiet -v -verbose -h -help] FILE [FILE..]")
}

func help() {
	printusage()
	fmt.Println()
	fmt.Println("Png2prg converts 320x200 images (png/gif/jpeg) to a c64 multicolor bitmap")
	fmt.Println("in koala format. It will find the best matching palette and backgroundcolor")
	fmt.Println("automatically, no need to modify your source images.")
	fmt.Println()
	fmt.Println("The resulting .prg includes the 2-byte start address and optional displayer.")
	fmt.Println()
	fmt.Println("  Bitmap: $2000 - $3f3f")
	fmt.Println("  Screen: $3f40 - $4327")
	fmt.Println("  D800:   $4328 - $470f")
	fmt.Println("  D021:   $4710")
	fmt.Println()
	fmt.Println("This tool can be used in all buildchains on most platforms.")
	fmt.Println("Multiple threads will be used when converting multiple images at once.")
	fmt.Println()
	fmt.Println("Options:")
	fmt.Println()
	flag.PrintDefaults()
	fmt.Println()
	os.Exit(0)
}

func main() {
	flag.Parse()
	srcfiles := flag.Args()
	if !quiet {
		fmt.Printf("png2prg %v by burglar\n", version)
	}
	if helpbool {
		help()
	}

	count := len(srcfiles)
	if count == 0 {
		printusage()
		os.Exit(0)
	}

	if display {
		koaladisplay, _ = base64.StdEncoding.DecodeString(koaladisplayb64)
	}

	c := make(chan int)
	for i := 0; i < count; i++ {
		go convertfile(c, srcfiles[i], i)
	}
	for i := 0; i < count; i++ {
		j := <-c
		if !quiet {
			destfilename := strings.TrimSuffix(srcfiles[j], filepath.Ext(srcfiles[j])) + ".prg"
			fmt.Printf("%-30v -> %v\n", srcfiles[j], destfilename)
		}
	}
}

func convertfile(c chan int, filename string, i int) {
	if verbose {
		log.Printf("running %v.%v %v\n", i, "-", time.Now().UnixNano())
	}
	m := getimagefromfile(filename)
	checkbounds(filename, m.Bounds())
	// process all colors and translate to c64 colors
	palettemap := getpalettemap(m)
	if verbose {
		log.Printf("palettemap len:%v %v\n", len(palettemap), palettemap)
	}

	// obtain background color
	bgcharcolorslice, screen := getbgcharcolorslice(m, palettemap)
	bgrgb, bgcolorindex := getbackgroundcolor(bgcharcolorslice)
	if verbose {
		log.Printf("background color: %v = %v\n", bgrgb, bgcolorindex)
	}

	// createkoala
	koala := getkoala(m, screen, bgrgb, bgcolorindex)
	// if verbose { log.Printf("koala: %v.%v %v : %v\n", i, "-", time.Now().UnixNano(), koala) }

	writefile(filename, koala, i)
	c <- i
}

func writefile(filename string, koala Koala, i int) {
	destfilename := strings.TrimSuffix(filename, filepath.Ext(filename)) + ".prg"
	if verbose {
		log.Printf("going to write file '%v'", destfilename)
	}
	f, err := os.Create(destfilename)
	check(err)
	defer f.Close()
	var n int
	if display {
		n, err = f.Write(koaladisplay)
		check(err)
		if verbose {
			log.Printf("wrote %d bytes to %v", n, destfilename)
		}
		paddinglength := 0x2000 - 0x7ff - len(koaladisplay)
		if verbose {
			log.Printf("paddinglength = %v", paddinglength)
		}
		padding := [8192]byte{}
		n, err = f.Write(padding[0:paddinglength])
		check(err)
		if verbose {
			log.Printf("wrote %d bytes to %v", n, destfilename)
		}

	} else {
		n, err = f.Write([]byte{0x00, 0x20})
		check(err)
		if verbose {
			log.Printf("wrote %d bytes to %v", n, destfilename)
		}
	}
	n, err = f.Write(koala.bitmap[:])
	check(err)
	if verbose {
		log.Printf("wrote %d bytes to %v", n, destfilename)
	}
	n, err = f.Write(koala.screencolor[:])
	check(err)
	if verbose {
		log.Printf("wrote %d bytes to %v", n, destfilename)
	}
	n, err = f.Write(koala.d800color[:])
	check(err)
	if verbose {
		log.Printf("wrote %d bytes to %v", n, destfilename)
	}
	n, err = f.Write(koala.bgcolor[:])
	check(err)
	if verbose {
		log.Printf("wrote %d bytes to %v", n, destfilename)
	}
	f.Sync()

	if verbose {
		log.Printf("file '%v' written successfully", destfilename)
	}
	if verbose {
		log.Printf("done %v.- %v\n", i, time.Now().UnixNano())
	}
}

func getkoala(m image.Image, screen Screen, bgrgb RGB, bgcolorindex byte) (koala Koala) {
	koala.bgcolor[0] = bgcolorindex
	for char := 0; char < 1000; char++ {
		bitmapindex := char * 8
		imagexindex := int(math.Mod(float64(char), 40)) * 8
		imageyindex := int(math.Floor(float64(char)/40)) * 8

		// we need colorindexes to know which bitpair is what, 00 is bg of course
		var colorindex1 = map[RGB]byte{bgrgb: 0}
		var colorindex2 = map[byte]byte{0: bgcolorindex}
		var bitpair byte = 1
		// TODO: predictable order of colors
		for rgb, colorindex := range screen.charcolors[char] {
			if bitpair > 4 {
				log.Fatalf("Too many colors in char %v.", char)
			}
			if colorindex == bgcolorindex {
				continue
			}
			colorindex1[rgb] = bitpair
			colorindex2[bitpair] = colorindex
			bitpair++
		}

		for byteindex := 0; byteindex < 8; byteindex++ {
			var bmpbyte byte = 0
			var bmppattern byte = 0
			for pixel := 0; pixel < 4; pixel++ {
				r, g, b, _ := m.At(imagexindex+(pixel*2), imageyindex+byteindex).RGBA()
				rgb := RGB{byte(r), byte(g), byte(b)}
				bmppattern = colorindex1[rgb]
				bmpbyte = bmpbyte | (bmppattern << (6 - (byte(pixel) * 2)))
			}
			koala.bitmap[bitmapindex+byteindex] = bmpbyte
		}

		if _, ok := colorindex2[1]; ok {
			koala.screencolor[char] = colorindex2[1] << 4
		}
		if _, ok := colorindex2[2]; ok {
			koala.screencolor[char] = koala.screencolor[char] | colorindex2[2]
		}
		if _, ok := colorindex2[3]; ok {
			koala.d800color[char] = colorindex2[3]
		}
	}
	return koala
}

func getbgcharcolorslice(m image.Image, palettemap map[RGB]byte) (bgcharcolorslice []map[RGB]byte,
	screen Screen) {
	for char := 0; char < 1000; char++ {
		charcolormap := getcolormapfromchar(char, m, palettemap)
		if len(charcolormap) > 4 {
			c64colcountmap := make(map[byte]byte, 16)
			for _, indexcolor := range charcolormap {
				c64colcountmap[indexcolor] = 1
			}
			if len(c64colcountmap) > 4 {
				log.Fatalf("FATAL ERROR: amount of colors in char %v > 4 : %v", char, c64colcountmap)
			}
		}
		screen.charcolors[char] = charcolormap
		if len(charcolormap) == 4 {
			bgcharcolorslice = append(bgcharcolorslice, charcolormap)
		}
	}

	return bgcharcolorslice, screen
}

func getbackgroundcolor(bgcharcolorslice []map[RGB]byte) (rgb RGB, colorindex byte) {
	// need to copy the map, as we delete stuff to eliminate false candidates
	var candidates = make(map[RGB]byte, 16)
	for key, val := range bgcharcolorslice[0] {
		candidates[key] = val
	}

Outer:
	for _, charcolormap := range bgcharcolorslice {
		for rgb, _ := range candidates {
			if _, ok := charcolormap[rgb]; !ok {
				delete(candidates, rgb)
				if len(candidates) == 1 {
					break Outer
				}
			}
		}
	}
	if verbose {
		log.Printf("candidates = %v\n", candidates)
	}

	for a, b := range candidates {
		rgb, colorindex = a, b
		break
	}
	return rgb, colorindex
}

func getcolormapfromchar(char int, m image.Image, palettemap map[RGB]byte) map[RGB]byte {

	x, y := getxyfromchar(char)
	// if verbose { log.Printf("char %v: x,y = %v,%v\n", char, x, y) }
	charcolormap := make(map[RGB]byte, 16)

	for j := 0; j < 32; j++ {
		pixelx, pixely := getxyfrompixel(j)
		r, g, b, _ := m.At(x+pixelx, y+pixely).RGBA()
		rgb := RGB{byte(r), byte(g), byte(b)}
		if _, ok := charcolormap[rgb]; !ok {
			charcolormap[rgb] = palettemap[rgb]
		}
		// if verbose { log.Printf("char %v: rgb = %v pixelx,pixely = %v,%v\n", char, rgb, x+pixelx, y+pixely) }
	}

	return charcolormap
}

func getxyfrompixel(i int) (x, y int) {
	return i << 1 & 7, i >> 2
}

func getxyfromchar(i int) (int, int) {
	return 8*i - (320 * int(math.Floor(float64(i/40)))),
		8 * int(math.Floor(float64(i/40)))
}

func getimagefromfile(filename string) image.Image {

	reader, err := os.Open(filename)
	check(err)
	defer reader.Close()

	m, _, err := image.Decode(reader)
	check(err)
	return m
}

func checkbounds(filename string, bounds image.Rectangle) {
	if (bounds.Min.X != 0) || (bounds.Min.Y != 0) ||
		(bounds.Max.X != 320) || (bounds.Max.Y != 200) {
		log.Printf("min x,y = %d,%d\n", bounds.Min.X, bounds.Min.Y)
		log.Printf("max x,y = %d,%d\n", bounds.Max.X, bounds.Max.Y)
		log.Fatalf("FATAL ERROR: image '%v' is not 320x200 pixels", filename)
	}
}

func getpalettemap(m image.Image) (palettemap map[RGB]byte) {
	bounds := m.Bounds()
	mindistance := 9e9
	var palettename string
	for name, palette := range c64palettes {
		curmap := make(map[RGB]byte, 16)
		totaldistance := 0.0
		for x := bounds.Min.X; x < bounds.Max.X; x += 2 {
			for y := bounds.Min.Y; y < bounds.Max.Y; y++ {

				r, g, b, _ := m.At(x, y).RGBA()
				rgb := RGB{byte(r), byte(g), byte(b)}

				if _, ok := curmap[rgb]; !ok {
					var distance float64
					curmap[rgb], distance = getclosestcolorindexwithpalette(rgb, palette)
					totaldistance += distance
				}
			}
		}
		if verbose {
			log.Printf("totaldistance: %v => %v\n", name, totaldistance)
		}
		if totaldistance < mindistance {
			palettemap = curmap
			palettename = name
			mindistance = totaldistance
		}
		if totaldistance == 0 {
			break
		}
	}
	if verbose {
		log.Printf("palette: %v\n", palettename)
	}

	return palettemap
}

func colordistance(c1 RGB, c2 RGB) float64 {
	cr := math.Abs(float64(c1.R) - float64(c2.R))
	cg := math.Abs(float64(c1.G) - float64(c2.G))
	cb := math.Abs(float64(c1.B) - float64(c2.B))
	// return math.Sqrt(float64((cr*cr)+(cg*cg)+(cb*cb)))
	return cr + cg + cb
}

func getclosestcolorindexwithpalette(color RGB, palette []C64RGB) (byte, float64) {
	distance := colordistance(color, palette[0].rgb)
	closestcolorindex := 0

	for i := 1; i < len(palette); i++ {
		d := colordistance(color, palette[i].rgb)
		if d < distance {
			distance = d
			closestcolorindex = i
		}
	}

	return byte(closestcolorindex), distance
}

func check(err error) {
	if err != nil {
		log.Fatal(err)
	}
}
